import {debounce} from './utils'

export enum DeviceScope {
  RECORD = 'scope.record',
  CAMERA = 'scope.camera'
}
export enum DeviceType {
    MIC = 'mic',
    CAMERA = 'camera'
}
/**
* 获取设备权限，查看当前是否有麦克风以及摄像头权限
* @param {Object} params - 参数对象
* @param {Boolean} params.microphone 麦克风权限
* @param {Boolean} params.camera     摄像头权限
*/
async function deviceCheck(types: Array<DeviceType>): Promise<boolean> {
  if (!types) return false
  const microphone = types.includes(DeviceType.MIC)
  const camera = types.includes(DeviceType.CAMERA)
  try {
    const res = await wx.getSetting()
    const isMicrophone = res.authSetting[DeviceScope.RECORD]
    const isCamera = res.authSetting[DeviceScope.CAMERA]
    // 同时检查麦克风和摄像头权限
    if (microphone && camera) {
      return isMicrophone && isCamera
    }
    // 只检查麦克风权限
    if (microphone) {
      return isMicrophone
    }
    // 只检查摄像头权限
    if (camera) {
      return isCamera
    }
    return false
  } catch (error) {
    return false
  }
}


function permissionPopup(): void {
  wx.showModal({
    title: '设备授权',
    content: '使用音视频功能请对相应设备进行授权',
    confirmText: '去设置',
    success: (res) => {
      if (res.confirm) {
        wx.openSetting()
      }
    },
  })
}
const debouncePermissionPopup = debounce(permissionPopup, 300)
export function deviceAuthorize(deviceType: DeviceType, scope: DeviceScope) {
  return function (target: any, key: string, descriptor: PropertyDescriptor): PropertyDescriptor {
    const originalMethod = descriptor.value
    descriptor.value = async function (...args: any[]): Promise<any> {
      if (!await deviceCheck([deviceType])) {
        try {
          await wx.authorize({scope})
        } catch (error) {
          debouncePermissionPopup()
          throw new Error(error)
        }
        return
      }
      // eslint-disable-next-line consistent-return
      return originalMethod.apply(this, args)
    }
    return descriptor
  }
}
